import torch
from deep_reps.similarity_functions import euclidean_distance, linear_kernel
from deep_reps.utils import compute_rsm, sqrtm_torch


def rsm_norm_difference(R, R_prime, similarity_function=linear_kernel):
    """RSM Norm Difference"""
    R = torch.tensor(R)
    R_prime = torch.tensor(R_prime)

    S = compute_rsm(R, similarity_function)
    S_prime = compute_rsm(R_prime, similarity_function)

    return float(euclidean_distance(S, S_prime))


def representation_similarity_analysis(R, R_prime, similarity_function=linear_kernel):
    """Representation Similarity Analysis"""
    R = torch.tensor(R)
    R_prime = torch.tensor(R_prime)

    S = compute_rsm(R, similarity_function)
    S_prime = compute_rsm(R_prime, similarity_function)

    n, _ = S.shape
    # The lower triangular part of the matrix (v \in \mathbb{R}^{n(n-1)/2})
    v = S[torch.tril_indices(n, n, offset=-1)].flatten()
    v_prime = S_prime[torch.tril_indices(n, n, offset=-1)].flatten()

    return float(torch.corrcoef(v, v_prime)[0, 1])


def centered_kernel_alignment(R, R_prime, similarity_function=linear_kernel):
    """Centered Kernel Alignment"""
    R = torch.tensor(R)
    R_prime = torch.tensor(R_prime)

    S = compute_rsm(R, similarity_function)
    S_prime = compute_rsm(R_prime, similarity_function)

    return HSIC(S, S_prime) / (HSIC(S, S) * HSIC(S_prime, S_prime)) ** 0.5


def centered_matrix(n):
    return torch.eye(n) - 1 / n * torch.ones((n, n)) @ torch.ones((n, n)).T


def HSIC(S, S_prime):
    n, _ = S.shape
    H = centered_matrix(n)
    return torch.trace(S @ H @ S_prime @ H) * 1 / (n - 1) ** 2


def mean_center(matrix):
    row_means = torch.mean(matrix, dim=1, keepdim=True)
    col_means = torch.mean(matrix, dim=0, keepdim=True)
    grand_mean = torch.mean(matrix)
    centered_matrix = matrix - row_means - col_means + grand_mean

    return centered_matrix


def dCov(S, S_prime):
    """Distance Covariance"""
    n = S.shape[0]
    S_hat = mean_center(S)
    S_prime_hat = mean_center(S_prime)

    return 1 / (n**2) * torch.sum(torch.sum(S_hat @ S_prime_hat, dim=1), dim=0)


def distance_correlation(R, R_prime, similarity_function=euclidean_distance):
    """Distance Correlation"""

    R = torch.tensor(R)
    R_prime = torch.tensor(R_prime)

    S = compute_rsm(R, similarity_function)
    S_prime = compute_rsm(R_prime, similarity_function)

    return float(
        dCov(S, S_prime) ** 2 / (dCov(S, S) ** 2 * dCov(S_prime, S_prime) ** 2) ** 0.5
    )


def make_positive_semi_definite(matrix, epsilon=1e-10):
    symmetric_matrix = (matrix + matrix.T) / 2
    psd_matrix = symmetric_matrix + epsilon * torch.eye(matrix.shape[0])
    return psd_matrix


def normalized_bures_similarity(R, R_prime, similarity_function=linear_kernel):
    """Normalized Bures Similarity"""
    R = torch.tensor(R)
    R_prime = torch.tensor(R_prime)

    S = compute_rsm(R, similarity_function)
    S = make_positive_semi_definite(S)
    S_prime = compute_rsm(R_prime, similarity_function)
    S_prime = make_positive_semi_definite(S_prime)
    S_sqrtm = sqrtm_torch(S)
    return float(
        torch.trace(S_sqrtm @ S_prime @ S_sqrtm) ** 0.5
        / (torch.trace(S) * torch.trace(S_prime)) ** 0.5
    )


def eigenspace_overlap_score(R, R_prime, similarity_function=linear_kernel):
    """Eigenspace Overlap Score"""
    R = torch.tensor(R)
    R_prime = torch.tensor(R_prime)

    S = compute_rsm(R, similarity_function)
    S_prime = compute_rsm(R_prime, similarity_function)

    _, eigmatrix = torch.linalg.eig(S)
    _, eigmatrix_prime = torch.linalg.eig(S_prime)

    denominator = torch.linalg.norm(eigmatrix.T @ eigmatrix_prime, "fro") ** 2
    divisor = max(eigmatrix.shape[1], eigmatrix_prime.shape[1])
    return float(denominator / divisor)


def get_s_minus_lambda(S, lamb=0):
    return torch.linalg.inv(S + lamb * torch.eye(S.shape[0]))


def gulp_similarity(R, R_prime):
    return 1 / R.shape[0] * (R.T @ R_prime)


def unified_linear_probing(R, R_prime, lamb=0.1):
    """Unified Linear Probing"""

    R = torch.tensor(R)
    R_prime = torch.tensor(R_prime)

    S = gulp_similarity(R, R)
    S_prime = gulp_similarity(R_prime, R_prime)
    S_r_r_prime = gulp_similarity(R, R_prime)

    S_minus_lambda = get_s_minus_lambda(S, lamb)
    S_prime_minus_lambda = get_s_minus_lambda(S_prime, lamb)

    trace_1 = torch.trace(S_minus_lambda @ S @ S_minus_lambda @ S)
    trace_2 = torch.trace(
        S_prime_minus_lambda @ S_prime @ S_prime_minus_lambda @ S_prime
    )
    trace_3 = torch.trace(
        S_minus_lambda @ S_r_r_prime @ S_prime_minus_lambda @ S_r_r_prime.T
    )

    return float((trace_1 + trace_2 - 2 * trace_3) ** 0.5)


def riemannian_similarity(R, R_prime):
    return 1 / R.shape[1] * (R @ R_prime.T)


def riemannian_distance(R, R_prime):
    """Riemannian Distance"""

    R = torch.tensor(R)
    R_prime = torch.tensor(R_prime)

    S = riemannian_similarity(R, R)
    S_prime = riemannian_similarity(R_prime, R_prime)

    S_inv = torch.linalg.inv(S)
    eigvalue, _ = torch.linalg.eig(S_inv @ S_prime)

    return float(torch.sum(torch.log(eigvalue) ** 2, dim=0) ** 0.5)



import torch


def pca(X, variance_retained=0.99):
    """Perform PCA on the data to retain the specified amount of variance."""
    X_centered = X - X.mean(dim=0)

    # Compute covariance matrix
    covariance_matrix = (X_centered.T @ X_centered) / (X_centered.size(0) - 1)

    # Compute eigenvalues and eigenvectors
    eigvals, eigvecs = torch.linalg.eigh(covariance_matrix)

    # Sort eigenvalues and eigenvectors in descending order
    sorted_indices = torch.argsort(eigvals, descending=True)
    eigvals = eigvals[sorted_indices]
    eigvecs = eigvecs[:, sorted_indices]

    # Compute the cumulative explained variance
    explained_variance = eigvals / eigvals.sum()
    cumulative_variance = torch.cumsum(explained_variance, dim=0)

    # Determine the number of components to retain
    num_components = (
        torch.searchsorted(cumulative_variance, variance_retained).item() + 1
    )

    # Project the data onto the principal components
    principal_components = eigvecs[:, :num_components]
    X_pca = X_centered @ principal_components

    return X_pca, principal_components


def vector_to_spd(vector, method="simple_spd"):
    # https://math.stackexchange.com/questions/3717983/generating-symmetric-positive-definite-matrix-from-random-vector-multiplication
    if method == "simple_spd":
        vector = torch.tensor(vector).reshape(-1, 1)
        spd_matrix = torch.mm(vector.T, vector)

    if method == "diagonal":
        vector = torch.tensor(vector).reshape(-1)
        D = torch.diag(vector)
        x, y = D.shape
        U = torch.ones((x, y))
        spd_matrix = D * U * D

    return spd_matrix


def sqrtm_torch(matrix):
    eigenvalues, eigenvectors = torch.linalg.eigh(matrix)
    sqrt_eigenvalues = torch.sqrt(eigenvalues)
    matrix_sqrt = eigenvectors @ torch.diag(sqrt_eigenvalues) @ eigenvectors.T
    return matrix_sqrt


def compute_rsm(input_tensor, similarity_function):
    n, _ = input_tensor.shape
    rsm_matrix = torch.zeros((n, n))
    for i in range(n):
        for j in range(n):
            rsm_matrix[i, j] = similarity_function(input_tensor[i], input_tensor[j])
    return rsm_matrix



import torch


def euclidean_distance(v1, v2):
    return torch.norm(v1 - v2)


def cosine_similarity(v1, v2):
    return torch.dot(v1, v2) / (torch.norm(v1) * torch.norm(v2))


def linear_kernel(v1, v2):
    return torch.dot(v1, v2) / (v1.size(0) - 1)


def rbf_kernel(v1, v2, sigma=0.1):
    return torch.exp(-(torch.norm(v1 - v2) ** 2) / (2 * sigma**2))


def pearson_correlation(v1, v2):
    v1_mean = torch.mean(v1)
    v2_mean = torch.mean(v2)
    covariance = torch.mean((v1 - v1_mean) * (v2 - v2_mean))
    v1_std = torch.std(v1)
    v2_std = torch.std(v2)
    return covariance / (v1_std * v2_std)


def rbf_kernel(X, Y, sigma=0.1):
    """Radial Basis Function Kernel"""
    return torch.exp(-sigma * torch.norm(X - Y) ** 2)



from deep_reps.cca_measures import (
    compute_cca,
    compute_standard_cca,
    compute_standard_svcca,
    compute_wvcca,
    compute_yanai_cca,
    compute_yanai_svcca,
)
from deep_reps.rsm_measures import (
    rsm_norm_difference,
    representation_similarity_analysis,
    centered_kernel_alignment,
    distance_correlation,
    normalized_bures_similarity,
    eigenspace_overlap_score,
    unified_linear_probing,
    riemannian_distance
)
from deep_reps.models import CLIPAndTokenizerLayers
from deep_reps.similarity_functions import (
    cosine_similarity,
    euclidean_distance,
    linear_kernel,
    pearson_correlation,
    rbf_kernel,
)
from deep_reps.utils import compute_rsm, pca, sqrtm_torch, vector_to_spd



from setuptools import find_packages, setup

VERSION = "0.0.3"
DESCRIPTION = "Representational Similarity Measures for Deep Learning"
LONG_DESCRIPTION = "DeepReps is a comprehensive library designed to centralize various measures of similarity for deep learning representations. It is specifically tailored for comparing the weights and outputs of deep learning models, providing a valuable toolkit for researchers and practitioners to assess and analyze model similarities."


setup(
    name="deep_reps",
    version=VERSION,
    author="Mezosky",
    author_email="<imezadelajara@gmail.com>",
    description=DESCRIPTION,
    packages=find_packages(),
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    install_requires=[
        "torch>=2.3.1",
        "matplotlib>=3.6.0",
        "numpy>=1.24.0",
        "pandas>=1.5.0",
        "plotly>=5.11.0",
    ],
    keywords=["python", "deep-learning", "torch", "similarity", "representations"],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
)



import open_clip
import torch
from PIL import Image


class CLIPAndTokenizerLayers(torch.nn.Module):
    def __init__(
        self, model_name="ViT-B-32", pretrained="laion400m_e32", device="cpu", seed=0
    ):
        super().__init__()
        self.model, self.preprocess, self.tokenizer = (
            open_clip.create_model_and_transforms(model_name, pretrained=pretrained)
        )
        self.model = self.model.to(device)
        self.tokenizer = open_clip.get_tokenizer(model_name)
        self.device = device
        self._set_seed(seed)
        self.model.eval()

        self.img_features = {}
        self.text_features = {}

        def get_activations(name, feature_dict):
            def hook(model, input, output):
                feature_dict[name] = output

            return hook

        for name, module in self.model.visual.named_modules():
            module.register_forward_hook(get_activations(name, self.img_features))

        for name, module in self.model.transformer.named_modules():
            module.register_forward_hook(get_activations(name, self.text_features))

    def _set_seed(self, seed):
        torch.manual_seed(seed)
        if self.device == "cuda":
            torch.cuda.manual_seed_all(seed)

    def _load_image_tensor(self, image_path):
        img_loaded = self.preprocess(Image.open(image_path)).unsqueeze(0)
        return img_loaded.to(self.device)

    def forward(self, img_path, text_tokens):
        img = self._load_image_tensor(img_path)
        with torch.no_grad():
            _ = self.model.encode_image(img)

        if text_tokens:
            text = self.tokenizer(text_tokens).to(self.device)
            with torch.no_grad():
                _ = self.model.encode_text(text)
            return self.img_features, self.text_features

        return self.img_features



# implementation in torch
import torch
from deep_reps.utils import pca


def mean_center(X):
    """Mean center the data, due to this method Assume mean-centered representations"""
    return X - X.mean(dim=0)


def compute_covariance_matrices(X, Y):
    n_samples = X.size(0)
    sigma_X = X.T @ X / (n_samples - 1)
    sigma_Y = Y.T @ Y / (n_samples - 1)
    sigma_XY = X.T @ Y / (n_samples - 1)
    return sigma_X, sigma_Y, sigma_XY


def compute_inverse_sqrt_matrix(matrix):
    eigvals, eigvecs = torch.linalg.eigh(matrix)
    eigvals = torch.clamp(eigvals, min=1e-10)  # Avoid division by zero
    inv_sqrt_matrix = eigvecs @ torch.diag(1.0 / torch.sqrt(eigvals)) @ eigvecs.T
    return inv_sqrt_matrix


def compute_cca(X, Y):
    # Center the data
    X_centered = mean_center(X)
    Y_centered = mean_center(Y)

    # Compute covariance matrices
    sigma_X, sigma_Y, sigma_XY = compute_covariance_matrices(X_centered, Y_centered)

    # Compute inverse square roots of covariance matrices
    sqrt_inv_sigma_X = compute_inverse_sqrt_matrix(sigma_X)
    sqrt_inv_sigma_Y = compute_inverse_sqrt_matrix(sigma_Y)

    # Compute the transformation matrix
    T = sqrt_inv_sigma_X @ sigma_XY @ sqrt_inv_sigma_Y

    # Perform SVD on the transformation matrix
    U, S, V = torch.svd(T)

    # Select the top output_dim components
    X_c = X_centered @ sqrt_inv_sigma_X @ U
    Y_c = Y_centered @ sqrt_inv_sigma_Y @ V

    return X_c, Y_c, S


def compute_standard_cca(X, Y):
    _, _, S = compute_cca(X, Y)
    return float(1 / X.size(1) * torch.sum(S))


def compute_yanai_cca(X, Y):
    _, _, S = compute_cca(X, Y)
    return float(1 / X.size(1) * torch.sum(S**2))


def compute_standard_svcca(X, Y, variance_retained=0.99):
    X_pca, _ = pca(X, variance_retained)
    Y_pca, _ = pca(Y, variance_retained)

    _, _, S = compute_cca(X_pca, Y_pca)
    return float(1 / X.size(1) * torch.sum(S))


def compute_yanai_svcca(X, Y, variance_retained=0.99):
    X_pca, _ = pca(X, variance_retained)
    Y_pca, _ = pca(Y, variance_retained)

    _, _, S = compute_cca(X_pca, Y_pca)
    return float(1 / X.size(1) * torch.sum(S**2))


def compute_wvcca(X, Y, output_dim=1):
    X_c, _, S = compute_cca(X, Y, output_dim)

    output_dim = X_c.size(1)
    X = mean_center(X)
    weights = torch.zeros(output_dim)
    for i in range(output_dim):
        Xw = X.T @ X_c[:, i]
        weights[i] = torch.abs(torch.sum(Xw @ X.T))

    # Normalize the weights
    weights = weights / torch.sum(weights)
    # Compute the final PWCCA measure
    return float(torch.sum(weights * S))
